--
-- Sending Twitter Tweets from a VLC
-- Base: http://regex.info/blog/lua/twitter
--
--
-- Copyright 2009 Jeffrey Friedl
-- (jfriedl@yahoo.com)
-- http://regex.info/blog/
--
-- Version 2 (June 14, 2009)
-- (version history follows at the end of this file)
-- Changed to fit VLC (December 10, 2010)
--
-- This requires my sha1.lua package (http://regex.info/blog/lua/sha1)
--
-- This exposes two public functions:
--
--    Twitter_AuthenticateNewCredentials()
--    Twitter_SendTweet(credential_bundle, status_text)
--
local function consumer_secret()   return "jH4Pvh03jC3VAwU5dcau3LFfiGdC8dL78aIFcqwU" end
local function consumer_key()      return "NtRc1d98zmCQmdyG11tscQ" end

--
-- Have this function return something unique to your application, such as a
-- hostname or reversed hostname, e.g. I use "regex.info/flickr" for my
-- upload-to-Flickr plugin. This is used only for generating a unique random
-- string, so the user will never see it.
--
local function string_unique_to_this_ap()   return "http://www.videolan.org/vlc/" end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------


local TwitterRequestTokenURL = 'http://api.twitter.com/oauth/request_token'
local TwitterAuthorizeURL    = 'http://api.twitter.com/oauth/authorize'
local TwitterAccessTokenURL  = 'http://api.twitter.com/oauth/access_token'

local base64 = require("base64")
require("hmac")
require("http")

local gtoken = nil
local gtoken_secret = nil

local function generate_nonce()
   return base64.enc(vlc.md5(vlc.md5(string_unique_to_this_ap()
                                  .. vlc.md5(tostring(math.random()) .. "random"))
                                  .. tostring(os.time())))
end

--
-- Like URL-encoding, but following OAuth's specific semantics
--
local function oauth_encode(str)
   return tostring(str):gsub('[^-%._~a-zA-Z0-9]', function(c)
        return string.format("%%%02x", c:byte()):upper()
    end)
end

--
-- Given a url endpoint, a GET/POST method, and a table of key/value args, build
-- the query string and sign it, returning the query string (in the case of a
-- POST) or, for a GET, the final url.
--
-- The args should also contain an 'oauth_token_secret' item, except for the
-- initial token request.
--
local function oauth_sign(url, method, args, onlysignature)

   assert(method == "GET" or method == "POST")

   local consumer_secret = consumer_secret()
   local token_secret    = args.oauth_token_secret or ""

   --
   -- Remove the token_secret from the args, 'cause we neither send nor sign it.
   -- (we use it for signing which is why we need it in the first place)
   --
   args.oauth_token_secret = nil

   -- Twitter does only HMAC-SHA1
   args.oauth_signature_method = 'HMAC-SHA1'


   --
   -- oauth-encode each key and value, and get them set up for a Lua table sort.
   --
   local keys_and_values = { }

   for key, val in pairs(args) do
      table.insert(keys_and_values,  {
                      key = oauth_encode(key),
                      val = oauth_encode(val)
                   })
   end

   --
   -- Sort by key first, then value
   --
   table.sort(keys_and_values, function(a,b)
                          if a.key < b.key then
                             return true
                          elseif a.key > b.key then
                             return false
                          else
                             return a.val < b.val
                          end
                       end)

   --
   -- Now combine key and value into key=value
   --
   local key_value_pairs = { }
   for _, rec in pairs(keys_and_values) do
      table.insert(key_value_pairs, rec.key .. "=" .. rec.val)
   end

   --
   -- Now we have the query string we use for signing, and, after we add the
   -- signature, for the final as well.
   --
   local query_string_except_signature = table.concat(key_value_pairs, "&")

   --

   -- Don't need it for Twitter, but if this routine is ever adapted for
   -- general OAuth signing, we may need to massage a version of the url to
   -- remove query elements, as described in http://oauth.net/core/1.0#rfc.section.9.1.2
   --
   -- More on signing:
   --   http://www.hueniverse.com/hueniverse/2008/10/beginners-gui-1.html
   --
   local SignatureBaseString = method .. '&' .. oauth_encode(url) .. '&' .. oauth_encode(query_string_except_signature)
   local key = oauth_encode(consumer_secret) .. '&' .. oauth_encode(token_secret)
   
   --
   -- Now have our text and key for HMAC-SHA1 signing
   --   
   local hmac_binary = hmac_sha1_binary(key, SignatureBaseString)

   --
   -- Base64 encode it
   --
   local hmac_b64 = base64.enc(hmac_binary)

   if onlysignature then
      return oauth_encode(hmac_b64)
   end
   
   --
   -- Now append the signature to end up with the final query string
   --
   local query_string = query_string_except_signature .. '&oauth_signature=' .. oauth_encode(hmac_b64)

   if method == "GET" then
      -- return the full url
      return url .. "?" .. query_string
   else
      -- for a post, just return the query string, so it can be included in the POST payload
      return query_string
   end
end

--
-- If an HTTP request returns nothing, check the headers and return some kind of reasonable
-- error message.
--
local function error_from_header(reply, headers)

   if not headers.status then
      return "Couldn't connect to twitter -- Internet connection down?"
   end
   
   local note = ""
   if headers.status ~= 401 then
      note = "Unexpected HTTP error reply from Twitter: " .. headers.status
   else
      note = "Authorization error (HTTP 401), please check PIN"
   end

   if reply then
      local error = reply:match("<error>(.-)</error>")
      if error then
         note = note .. ": " .. error
      end
   end

   return note
end

--
-- Start a sequence that allows the user to authenticate their Twitter account
-- to the plugin. This can't be run on the main LR task, so be sure it's downwind
-- of a LrTask.startAsyncTask() or LrFunctionContext.postAsyncTaskWithContext().
--
-- On failure, it returns nil and an error message.
--
-- On success, it returns a "credential bundle" table along the lines of:
--       
--       {
--          oauth_token        = "jahdhYHajdkajaeh"
--          oauth_token_secret = "GFWFGN$7gIN9Nf8huN&G^G#736nx7N&ZY#SyZz",
--          user_id            = "14235768",
--          screen_name        = "jfriedl",
--       }
--
-- You should cache this credential-bundle table somewhere (e.g. in the
-- Lightroom Prefs) and use it for subsequent interaction with Twitter on behalf
-- of the user, forever, unless attempting to use it results in an error
-- (at which point you probably want to uncache it).
-- 
function Twitter_AuthenticateNewCredentials()
   --
   -- First ping Twitter to get a request token.
   --
   local url = oauth_sign(TwitterRequestTokenURL,
                          "GET",
                          {
                             oauth_consumer_key = consumer_key(),
                             oauth_timestamp    = os.time(),
                             oauth_version      = '1.0',
                             oauth_callback     = "oob",
                             oauth_nonce        = generate_nonce(),
                          })

   local result, headers, errors = http_get(url)

   if errors then
      return nil, errors
   end
   if not result or headers.status ~= 200 then
      return nil, error_from_header(result, headers)
   end

   gtoken        = result:match('oauth_token=([^&]+)')
   gtoken_secret = result:match('oauth_token_secret=([^&]+)')

   if not gtoken then
      return nil, "Couldn't get request token from Twitter"
   end

   --
   -- Tell the user that they'll have to permission their account to allow this
   -- app to have access, and give them a chance to bail.
   --
   local url = TwitterAuthorizeURL .. '?oauth_token=' .. oauth_encode(gtoken)
   return url, nil
end

function Twitter_AuthenticateNewCredentials_PIN(PIN)
   if not PIN then
      return nil, "canceled"
   end

   --
   -- Now that your app should have permission, go to Twitter and get the
   -- authentication token that will let you interact with Twitter on behalf of the
   -- user.
   --
   local url = oauth_sign(TwitterAccessTokenURL,
                          "GET",
                          {
                             oauth_consumer_key = consumer_key(),
                             oauth_timestamp    = os.time(),
                             oauth_version      = '1.0',
                             oauth_callback     = "oob",
                             oauth_nonce        = generate_nonce(),
                             oauth_token        = gtoken,
                             oauth_token_secret = gtoken_secret,
                             oauth_verifier     = PIN,
                          })

   local result, headers, errors = http_get(url)

   if errors then
      return nil, errors
   end
   if not result or headers.status ~= 200 then
      return nil, error_from_header(result, headers)
   end

   local oauth_token        = result:match(       'oauth_token=([^&]+)')
   local oauth_token_secret = result:match('oauth_token_secret=([^&]+)')
   local user_id            = result:match(           'user_id=([^&]+)')
   local screen_name        = result:match(       'screen_name=([^&]+)')
   
   if oauth_token and oauth_token_secret and user_id and screen_name then
      --
      -- Got it
      --
	  return {
         oauth_token        = oauth_token,
         oauth_token_secret = oauth_token_secret,
         user_id            = user_id,
         screen_name        = screen_name,
      }
   end

   return nil, "Unexpected reply from Twitter:" .. result

end

--
-- Twitter_SendTweet(credential_bundle, status_text)
--
-- Given a credential bundle (as returned by Twitter_AuthenticateNewCredentials),
-- and the text of a new tweet, send it.
--
-- Returns true on success, nil if the user has revoked permission for your app
-- (and thus the credential bundle should be discarded and no longer cached),
-- and false on other failure (e.g. network is down)
--
function Twitter_SendTweet(credential_bundle, status_text)

   assert(type(status_text)                          == 'string')
   assert(type(credential_bundle)                    == 'table')
   assert(type(credential_bundle.oauth_token)        == 'string')
   assert(type(credential_bundle.oauth_token_secret) == 'string')

   local url = "http://api.twitter.com/1/statuses/update.xml"
   local ostime = os.time()
   local nonce = generate_nonce()

   -- http://apiwiki.twitter.com/Twitter-REST-API-Method%3A-statuses%C2%A0update
   local signature = oauth_sign(url,
                                   "POST",
                                   {
                                      status             = status_text,
                                      oauth_consumer_key = consumer_key(),
                                      oauth_timestamp    = ostime,
                                      oauth_version      = '1.0',
                                      oauth_nonce        = nonce,
                                      oauth_token        = credential_bundle.oauth_token,
                                      oauth_token_secret = credential_bundle.oauth_token_secret,
                                   },
								   true)

   --
   -- Twitter requires the Content-Type and Content-Length be set, or they refuse the authentication.
   --
   local postdata = "status=" .. oauth_encode(status_text)
   local result, headers, errors = http_post(url,
                                       postdata,
                                       {
                                          {
                                             field = 'Content-Type',
                                             value = 'application/x-www-form-urlencoded',
                                          },
                                          {
                                             field = 'Content-Length',
                                             value = tostring(#postdata)
                                          },
										  {
										     field = 'Authorization',
											 value = "OAuth "
											 .. "oauth_nonce=\"" .. oauth_encode(nonce) .. "\", "
											 .. "oauth_signature_method=\"HMAC-SHA1\", "
											 .. "oauth_timestamp=\"" .. ostime .. "\", "
											 .. "oauth_consumer_key=\"" .. oauth_encode(consumer_key()) .. "\", "
											 .. "oauth_token=\"" .. oauth_encode(credential_bundle.oauth_token) .. "\", "
											 .. "oauth_signature=\"" .. signature .. "\", "
											 .. "oauth_version=\"1.0\""
										  }
                                       })
	if errors then
		return nil, errors
	end
   if not result then
      return nil, error_from_header(headers)
   end

   if result:match("Failed to validate") then
      return nil -- user revoked permission
   end
   
   -- For lack of a better idea, I'll pick the presense of "<created_at>" as a signal of success.
   if result:match('<created_at>') then
      return true
   else
      return false, result:match('<error>(.*)</error>') -- some kind of failure... unknown response
   end
end
